import time
import requests
from base64 import b64encode
from typing import List
from urllib.parse import urlencode
from .base import Plugin
from ..models import SimpleTrack
from ..exceptions import MusicBrowserException


class SpotifyAuthentication:
    """
    Manage token based authentication for Spotify Plugin.
    """

    def __init__(self, client_id, client_secret):
        # workout base64 header required by spotify to authenticate using client credentials.
        self.basic_token = b64encode(("%s:%s" % (client_id, client_secret)).encode("utf-8")).decode("utf-8")
        # token lazy loading. Indicates that a new token must be issued from now.
        self.expires = time.time()

    def authenticate(self):
        # exchange client credentials for a token against spotify accounts API.
        response = requests.post(
            "https://accounts.spotify.com/api/token",
            data={"grant_type": "client_credentials"},
            headers={
                "Authorization": "Basic %s" % self.basic_token,
                "Content-Type": "application/x-www-form-urlencoded",
            },
        )
        # ensure that proper token response has been received
        if response.status_code != 200:
            raise MusicBrowserException("Unable to retrieve access_token from Spotify API.")

        token_data = response.json()
        # store useful token data as properties
        self.access_token = token_data["access_token"]
        self.expires = time.time() + token_data["expires_in"]

    def has_expired(self):
        # check if expiration time is passed
        return self.expires < time.time()

    def get_token(self):
        # ensure that known token is up-to-date
        if self.has_expired():
            self.authenticate()
        # ... then return valid access token
        return self.access_token


class SpotifyPlugin(Plugin):
    """
    Define methods to search tracks on spotify
    and complete metadata from spotify's database.
    """

    def __init__(self, client_id, client_secret):
        self.auth = SpotifyAuthentication(client_id, client_secret)

    @property
    def known_fields(self):
        return {"title", "artist", "album"}

    def query(self, query_params, query_fields):
        pass

    def search(self, query) -> List[SimpleTrack]:
        r = requests.get(
            "https://api.spotify.com/v1/search?" + urlencode({"type": "track", "query": query}),
            headers={"Authorization": "Bearer %s" % self.auth.get_token()},
        )

        tracks = []

        for track_data in r.json()["tracks"]["items"]:
            tracks.append(
                SimpleTrack(
                    album=track_data["album"]["name"],
                    artist=track_data["artists"][0]["name"],
                    title=track_data["name"],
                    cover=track_data["album"]["images"][0]["url"],
                    source={"id": track_data["id"], "platform": "spotify"},
                )
            )

        return tracks



import abc
from typing import List, Set
from ..models import SimpleTrack


class Plugin(abc.ABC):
    @abc.abstractproperty
    def known_fields(self) -> Set[str]:
        raise NotImplementedError("You must implement known_fields property of your Plugin")

    @abc.abstractmethod
    def query(self, query_params, query_fields) -> dict:
        raise NotImplementedError("You must implement query method of your Plugin")

    @abc.abstractmethod
    def search(self, query: str) -> List[SimpleTrack]:
        raise NotImplementedError("You must implement search method to use a plugin as search_plugin...")



import json


class MusicBrowserSerializer(json.JSONEncoder):
    def default(self, obj):
        try:
            return obj.__json__()
        except AttributeError:
            return json.JSONEncoder.default(self, obj)



from typing import List
from .plugins import Plugin
import music_browser.exceptions

REQUIRED = True
OPTIONAL = False


class TrackMetadataQuery:
    """
    Object used to retrieve requested track metadata, given a few info about the track.
    """

    def __init__(self, plugins: List[Plugin], known_fields=List[str]):
        self.plugins = plugins
        self.known_fields = known_fields

    def query(
        self,
        query_string: str = None,
        title: str = None,
        artist: str = None,
        spotify_id: str = None,
    ):
        """
        Specify known info about tracks that can be used to perform search.
        """
        self.query_string = query_string
        self.spotify_id = spotify_id
        return self

    def fields(self, **query_fields):
        """
        Specify fields that you want to retrieve from this query.

        Keyword Arguments:
            **:
                key -- (str) Name of the field to retrieve
                value -- (bool) Indicates that this field is mandatory (required)
        """
        for query_field, required in query_fields.items():
            if required and not hasattr(self.known_fields, query_field):
                raise music_browser.exceptions.UnknownField("No plugin can resolve required field %s." % query_field)
        self.query_fields = query_fields
        return self

    def result(self):
        """
        Try to fetch a track's requested metadata using plugins.
        Each plugin can partially complete fields of the result.
        For each field, first plugin that returns the field win.
        When all required fields are present in result, returns.
        """

        raise NotImplementedError()



import requests
from typing import List
from urllib.parse import urlencode
from .base import Plugin
from ..models import SimpleTrack


class DeezerPlugin(Plugin):
    @property
    def known_fields(self):
        return set()

    def query(self):
        pass

    def search(self, query) -> List[SimpleTrack]:
        response = requests.get("https://api.deezer.com/search?" + urlencode({"q": query}))

        tracks = []
        for track_data in response.json()["data"]:
            tracks.append(
                SimpleTrack(
                    album=track_data["album"]["title"],
                    artist=track_data["artist"]["name"],
                    title=track_data["title"],
                    cover="https://e-cdns-images.dzcdn.net/images/cover/%s/264x264-000000-80-0-0.jpg"
                    % track_data["md5_image"],
                    source={"id": str(track_data["id"]), "platform": "deezer"},
                )
            )

        return tracks



import setuptools


def get_description_from_readme():
    with open("README.md") as readme:
        return readme.read()


setuptools.setup(
    name="music-browser",
    description="Music search engine for CrossPlay web service.",
    author="Sylvan Le Deunff",
    author_email="sledeunf@gmail.com",
    version="0.0.4",
    url="https://github.com/tms-studio/python-music-browser",
    long_description=get_description_from_readme(),
    long_description_content_type="text/markdown",
    license_file="LICENSE",
    packages=setuptools.find_packages(),
    install_requires=["marshmallow", "requests"],
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
    ],
)



from typing import List
from .models import SimpleTrack
from .metadata import TrackMetadataQuery


class MusicBrowser:
    """
    Object that abstract plugins usage to query tracks metadata from multiple sources.

    Keywords Arguments:
        plugins -- (list) Plugins ordered by priority used to fetch metadata.
    """

    def __init__(self, search_plugin=None, completion_plugins=None):
        self.search_plugin = search_plugin
        self.plugins = []
        self.known_fields = set()

        if completion_plugins:
            # register plugins that improve MusicBrowser sources
            for plugin_module in completion_plugins:
                self.use_plugin(plugin_module)

    def query(self, query_string=None, title=None, artist=None, spotify_id=None):
        """
        Keywords Arguments:
            query_string -- (str) What you would type in a browser to search this track.
            title -- (str) Title of the track.
            artist -- (str) Name of the artist who created this track.

        Returns:
            track_query -- An object used to do lazy loading of tracks metadata.
        """
        return TrackMetadataQuery(self.plugins).query(
            query_string=query_string, title=title, artist=artist, spotify_id=spotify_id
        )

    def search(self, query: str) -> List[SimpleTrack]:
        """
        Use search_plugin to retrieve a list of tracks.
        """
        return self.search_plugin.search(query)

    def use_plugin(self, plugin):
        """
        Declare the use of the plugin.
        """
        self.plugins.append(plugin)

        for field in plugin.known_fields:
            self.known_fields.add(field)



class Source:
    def __init__(self, id: str, platform: str):
        self.id = id
        self.platform = platform

    def __json__(self):
        return {"id": self.id, "platform": self.platform}


class SimpleTrack:
    def __init__(self, title: str, artist: str, album: str, cover: str, source: Source):
        self.title = title
        self.artist = artist
        self.album = album
        self.cover = cover
        self.source = source

    def __str__(self):
        max_width = 20
        return "SimpleTrack:   %s   %s   %s" % (
            (self.title if len(self.title) < max_width else self.title[: max_width - 3] + "...").ljust(max_width),
            (self.artist if len(self.artist) < max_width else self.artist[: max_width - 3] + "...").ljust(max_width),
            (self.album if len(self.album) < max_width else self.album[: max_width - 3] + "...").ljust(max_width),
        )

    def __json__(self):
        return {
            "title": self.title,
            "artist": self.artist,
            "album": self.album,
            "cover": self.cover,
            "source": self.source,
        }
